/*-
 * Copyright (C)2010..2014 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

static const char copyright[] = "\
@(#)Copyright (C)2010..2014 @BABOLO http://www.babolo.ru/\n\
@(#)All rights reserved.\n\
@(#)V.M "VERS;
static const char rcsid[] = "$Id: blin.c,v 1.22 2014/09/22 14:26:21 babolo Exp $";

#define BLIN_COMPAT VMAJOR

#include <sys/types.h>
#include <sysexits.h>
#include <strings.h>
#include <string.h>
#include <syslog.h>
#include <stdlib.h>
#include <stdarg.h>
#include <stdio.h>
#include <errno.h>
#include <time.h>
#include <err.h>
#include "../BLINflag.h"

#define BLIN_MAININI   0x000001
#define BLIN_FLEGAL    (~BLIN_MAININI)
#define BLIN_MIMASK(A) (((u_int64_t)1 << A) - 1)
#define ERROUT(A,B)    do {ex = (A); errno = (B); goto out;} while(0)
#define BLIN_MAXTXTSTA 30
#define BLIN_MAXTXTDUM 88
#define BLIN_MAXSTRDUM 15
#define BLIN_MAXCRY1   256

blin_pamparam blin_pamparam_internal =
{ 0x00000000, 0
, { {NULL, BLIN_MODWARN | BLIN_MODEOAU, LOG_LOCAL7 | LOG_ERR}
  , {NULL, BLIN_MODWARN | BLIN_MODEOAU, LOG_LOCAL7 | LOG_WARNING}
  , {NULL, BLIN_MODPRNT | BLIN_MODEOAU, LOG_LOCAL7 | LOG_DEBUG}
  , {NULL, BLIN_MODPRNT | BLIN_MODEOAU, LOG_LOCAL7 | LOG_DEBUG}
  , {NULL, BLIN_MODPRNT | BLIN_MODEOAU, LOG_LOCAL7 | LOG_DEBUG}
  , {NULL, BLIN_MODPRNT | BLIN_MODEOAU, LOG_LOCAL7 | LOG_DEBUG}
  , {NULL, BLIN_MODPRNT | BLIN_MODEOAU, LOG_LOCAL7 | LOG_DEBUG}
  , {NULL, BLIN_MODPRNT | BLIN_MODEOAU, LOG_LOCAL7 | LOG_DEBUG}
} };

__weak_reference(blin_pamparam_internal, blin_internal_flags);
__weak_reference(blin_pamparam_internal, blin_direct_flags);

static const char * const
fac[] =
{ "KERN"  , "USER"  , "MAIL"    , "DAEMON", "AUTH"  , "SYSLOG"  , "LPR"    , "NEWS"
, "UUCP"  , "CRON"  , "AUTHPRIV", "FTP"   , "NTP"   , "SECURITY", "CONSOLE", "reserved"
, "LOCAL0", "LOCAL1", "LOCAL2"  , "LOCAL3", "LOCAL4", "LOCAL5"  , "LOCAL6" , "LOCAL7"
};

static const char * const
pri[] = {"EMERG", "ALERT", "CRIT", "ERR", "WARNING", "NOTICE", "INFO", "DEBUG"};

static const char * const
blinf[] = 
{ "MAININI", NULL  , NULL  , NULL  , NULL  , NULL  , NULL  , NULL
, NULL     , NULL  , NULL  , NULL  , NULL  , NULL  , NULL  , NULL
, NULL     , NULL  , NULL  , NULL  , NULL  , NULL  , NULL  , "TRAC"
, "BIT0"   , "BIT1", "BIT2", "BIT3", "BIT4", "BIT5", "BIT6", "BIT7"
};

static const char * const
whatf[] = 
{ NULL     , NULL     , NULL     , NULL  , "MODDTXT", NULL     , NULL     , NULL
, NULL     , NULL     , NULL     , NULL  , "MODDSTR", "MODDNON", "MODDREL", "MODTIME"
, "MODEOLF", "MODEOCR", "MODEOAU", NULL  , NULL     , "MODLOGR", "MODPRNT", "MODWARN"
, NULL     , NULL     , NULL     , NULL  , NULL     , NULL     , NULL     , NULL
};

#define SPC0 64
static const char *
spaces = "                                                                "; /* SPC0 * ' ' */

size_t
blin_internal_strnlen(const char *s, size_t maxlen) {
    size_t len;

    for (len = 0; len < maxlen; len++, s++) {
        if (!*s) break;
    }
    return (len);
}
__weak_reference(blin_internal_strnlen, strnlen);

static void
/*****************************************************************************************************/
blin_init() {                                                                                    /****
 *****************************************************************************************************/
    int i;

    if  (!(blin_pamparam_internal.flags & BLIN_MAININI)) {
        if  (blin_pamparam_internal.flags & BLIN_TRAC) {
            fprintf(stderr, "BLIN: +blin_init %08X\n", blin_pamparam_internal.flags);
        }
        for (i = 0; i < 8; i++) blin_pamparam_internal.w[i].file = stderr;
        blin_pamparam_internal.flags |= BLIN_MAININI;
        if  (blin_pamparam_internal.flags & BLIN_TRAC) {
            fprintf(stderr, "BLIN: -blin_init %08X\n", blin_pamparam_internal.flags);
}   }   }

static u_int32_t
/*****************************************************************************************************/
blin_do(u_int32_t mode, u_int32_t w, u_int32_t legal, u_int32_t *f) {                            /****
 *****************************************************************************************************/
    u_int32_t r = *f;

    if  (blin_pamparam_internal.flags & BLIN_TRAC) {
        fprintf(stderr, "BLIN: +blin_do %08X %08X %08X %08X\n", mode, w, legal, *f);
    }
    if  (!(mode & BLIN_CTL_RISE)) {
        w = (mode & BLIN_CTL_NPRM) ? 0 : w;
    } else {
        w = (((((mode & BLIN_CTL_NPRM) ? r : w) << 1) & BLIN_MASK) | BLIN_BIT0);
    }
    if  (mode & BLIN_CTL_FNEG) w = ~w;
    w &= legal;
    switch (mode & (BLIN_CTL_NONE | BLIN_CTL_FAFM)) {
    case BLIN_CTL_FSET: *f |= w;                    break;
    case BLIN_CTL_FRST: *f &= ~w;                   break;
    case BLIN_CTL_FXOR: *f ^= w;                    break;
    case BLIN_CTL_FEQU: *f = (*f & ~legal) | w;     break;
    }
    if  (mode & BLIN_CTL_NPRE) r = *f;
    if  (mode & BLIN_CTL_RMSK) r &= legal;
    if  (blin_pamparam_internal.flags & BLIN_TRAC) {
        fprintf(stderr, "BLIN: -blin_do %08X\n", r);
    }
    return(r);
}

int
/*****************************************************************************************************
 *****************************************************************************************************/
blin_64(u_int64_t flags, const char * const *nm, int x, FILE *fo) {                              /****
 *****************************************************************************************************
 *****************************************************************************************************/
    int i, k, z, ex = 0;
    u_int64_t m;

    if  ((32 < x) && (flags & 0xFFFFFFFF00000000ULL)) {
        ex += fprintf(fo, "%016"BLIN_O"X", flags);
    } else {
        ex += fprintf(fo, "%08X", (u_int32_t)flags);
    }
    m = 1;
    z = x;
    i = 0;
    if  (0 > x) {
        z = -x;
        i = x;
    }
    for (k = 0; !!m && (i < z); ++i, m <<= 1) {
        if ((flags & m) && ((0 > i) || !nm[i])) {
            ex += fprintf(fo, "%sunkn", k++ ? "," : "<");
            break;
    }   }
    m = 1;
    if  (0 > x) m <<= z;
    for (i = 0; flags && (i < z); ++i, m <<= 1) {
        if  ((flags & m) && !!nm[i]) {
            ex += fprintf(fo, "%s%s", k++ ? "," : "<", nm[i]);
    }   }
    ex += fprintf(fo, ">");
    return(ex);
}

void *
/*****************************************************************************************************
 *****************************************************************************************************/
blin_ctl(u_int32_t mode, ...) {                                                                  /****
 *****************************************************************************************************
 *****************************************************************************************************/
    void *res = NULL;
    u_int32_t m;
    va_list ap;
    int i, d;

    va_start(ap, mode);
    if  (blin_pamparam_internal.flags & BLIN_TRAC) {
        fprintf(stderr, "BLIN: +blin_ctl %08X\n", mode);
    }
    blin_init();
    switch (mode & BLIN_CTL_OPER) {
    case BLIN_CTL_FLAG: {
            u_int32_t w = va_arg(ap, u_int32_t);

            res = (void *)(size_t)blin_do(mode, w, BLIN_FLEGAL, &blin_pamparam_internal.flags);
        }
        break;
    case BLIN_CTL_WHAT: {
            u_int32_t w = va_arg(ap, u_int32_t);

            for (m = BLIN_BIT0, i = 0; m; m <<= 1, ++i) {
                if  (mode & m) {
                    res = (void *)(size_t)blin_do(mode, w, ~0, &blin_pamparam_internal.w[i].what);
        }   }   }
        break;
    case BLIN_CTL_FNAM: {
            char *f = va_arg(ap, char *);
            FILE *w = fopen(f, "a");

            for (m = BLIN_BIT0, i = 0; m < BLIN_MASK; m <<= 1, ++i) {
                if  (mode & m) {
                    res = blin_pamparam_internal.w[i].file;
                    if  (!(mode & BLIN_CTL_NONE)) blin_pamparam_internal.w[i].file = w;
        }   }   }
        break;
    case BLIN_CTL_FDES: {
            int f = va_arg(ap, int);
            FILE *w = fdopen(f, "a");

            for (m = BLIN_BIT0, i = 0; m < BLIN_MASK; m <<= 1, ++i) {
                if  (mode & m) {
                    res = blin_pamparam_internal.w[i].file;
                    if  (!(mode & BLIN_CTL_NONE)) blin_pamparam_internal.w[i].file = w;
        }   }   }
        break;
    case BLIN_CTL_FILE: {
            void *w = va_arg(ap, void *);

            for (m = BLIN_BIT0, i = 0; m < BLIN_MASK; m <<= 1, ++i) {
                if  (mode & m) {
                    res = blin_pamparam_internal.w[i].file;
                    if  (!(mode & BLIN_CTL_NONE)) blin_pamparam_internal.w[i].file = w;
        }   }   }
        break;
    case BLIN_CTL_PRIO: {
            u_int32_t w = va_arg(ap, u_int32_t);

            for (m = BLIN_BIT0, i = 0; m < BLIN_MASK; m <<= 1, ++i) {
                if  (mode & m) {
                    res = (void *)(ssize_t)(blin_pamparam_internal.w[i].prio);
                    if  (!(mode & BLIN_CTL_NONE)) blin_pamparam_internal.w[i].prio = w;
        }   }   }
        break;
    case BLIN_CTL_CPRI: {
            res = (void *)copyright;
        }
        break;
    case BLIN_CTL_RCSI: {
            res = (void *)rcsid;
        }
        break;
    case BLIN_CTL_VERS: {
            res = (void *)VERS;
        }
        break;
    case BLIN_CTL_DUMP: {
            int k;

        ;   fprintf( stderr, "blin_pamparam_internal @ %"BLIN_X" "
                   , BLIN_I(&blin_pamparam_internal)
                   );
        ;   blin_64(blin_pamparam_internal.flags, blinf, 32, stderr);
        ;   if  (blin_pamparam_internal.err) {
        ;       fprintf(stderr, " !!! %d !!!\n", blin_pamparam_internal.err);
        ;   } else {
        ;       fprintf(stderr, "\n");
        ;   }
        ;   for (i = 0; i < 8; ++i) {
#       define  SPC1 40
#       define  SPC2 24
        ;       fprintf(stderr, "  %u ", i);
        ;       d = blin_64(blin_pamparam_internal.w[i].what, whatf, 32, stderr);
        ;       if  (SPC1 > d) {
        ;           fprintf(stderr, "%s", &spaces[SPC0 - SPC1 + d]);
        ;           d = 0;
        ;       } else {
        ;           d -= SPC1;
        ;       }
        ;       k = 0;
        ;       m = blin_pamparam_internal.w[i].prio;
        ;       d += fprintf(stderr, " %08X", m);
        ;       if  (m & ~(LOG_PRIMASK | LOG_FACMASK)) {
        ;           d += fprintf(stderr, "%sunkn", k++ ? "," : "<");
        ;       }
        ;       if  (((m & LOG_FACMASK) >> 3) >= LOG_NFACILITIES) {
        ;           d += fprintf(stderr, "%sbig", k++ ? "," : "<");
        ;       }
        ;       d += fprintf(stderr, "%s%s", k++ ? "," : "<", fac[(m & LOG_FACMASK) >> 3]);
        ;       d += fprintf(stderr, "%s%s", k++ ? "," : "<", pri[m & LOG_PRIMASK]);
        ;       d += fprintf(stderr, ">");
        ;       if  (SPC2 > d) fprintf(stderr, "%s", &spaces[SPC0 - SPC2 + d]);
        ;       if  (blin_pamparam_internal.w[i].file == stdin) {
        ;           fprintf(stderr, " file=stdin\n");
        ;       } else if (blin_pamparam_internal.w[i].file == stdout) {
        ;           fprintf(stderr, " file=stdout\n");
        ;       } else if (blin_pamparam_internal.w[i].file == stderr) {
        ;           fprintf(stderr, " file=stderr\n");
        ;       } else {
        ;           fprintf(stderr, " file=%"BLIN_X"\n", BLIN_I(blin_pamparam_internal.w[i].file));
    }   }   }   }
    if  (blin_pamparam_internal.flags & BLIN_TRAC) {
        fprintf(stderr, "BLIN: -blin_ctl %08X\n", (u_int32_t)res);
    }
    va_end(ap);
    return(res);
}

static void
/*****************************************************************************************************/
doout(int level, char *q) {                                                                      /****
 *****************************************************************************************************/
    u_int32_t ctl = blin_pamparam_internal.w[level & 7].what;

    if  (ctl & BLIN_MODWARN) {
        warnx("%s", q);
    }
    if  (ctl & BLIN_MODPRNT) {
        fprintf(blin_pamparam_internal.w[level & 7].file, "%s", q);
    }
    if  (ctl & BLIN_MODLOGR) {
        syslog(blin_pamparam_internal.w[level & 7].prio, "%s", q);
}   }

static void
/*****************************************************************************************************/
dotime(int level) {                                                                              /****
 *****************************************************************************************************/
    u_int32_t ctl = blin_pamparam_internal.w[level].what;
    struct timespec tp;

    if  (ctl & BLIN_MODTIME) {
        if  (0 > clock_gettime(CLOCK_REALTIME, &tp)) {
            ctl &= ~BLIN_MODTIME;
        } else {
            if  (ctl & BLIN_MODWARN) {
                warnx("%10"BLIN_U".%09lu", tp.tv_sec, tp.tv_nsec);
            }
            if  (ctl & BLIN_MODPRNT) {
                fprintf( blin_pamparam_internal.w[level].file, "%10"BLIN_U".%09lu\n"
                       , tp.tv_sec, tp.tv_nsec
                       );
            }
            if  (ctl & BLIN_MODLOGR) {
                syslog( blin_pamparam_internal.w[level].prio, "%10"BLIN_U".%09lu"
                      , tp.tv_sec, tp.tv_nsec
                      );
}   }   }   }

int
/*****************************************************************************************************
 *****************************************************************************************************/
blin_cry1(int ctl, const char *format, ...) {                                                    /****
 *****************************************************************************************************
 *****************************************************************************************************/
    char qq[BLIN_MAXCRY1 + 4], *buf, *ptr;
    ssize_t avail, t, need = 0;
    struct timespec tp;
    u_int32_t control;
    int er, ex = EX_OK;
    va_list ap;

    er = errno;
    blin_init();
    if  (blin_pamparam_internal.flags & BLIN_TRAC) {
        fprintf(stderr, "BLIN: +blin_cry1 %08X\n", ctl);
    }
    control = blin_pamparam_internal.w[ctl & 7].what;
    if  (control & BLIN_MODTIME) {
        if  (0 > clock_gettime(CLOCK_REALTIME, &tp)) control &= ~BLIN_MODTIME;
    }
    need = 0;
    do {
        va_start(ap, format);
        if  (!need) {
            buf = qq;
            avail = BLIN_MAXCRY1;
        } else {
            buf = malloc(need + 5);
            avail = need + 1;
            need = 0;
            if  (!buf) {
                ifBLIN_QX0("malloc");
                ERROUT(EX_OSERR, ENOMEM);
        }   }
        ptr = buf;
        if  (control & BLIN_MODTIME) {
            t = snprintf(ptr, avail + 1, "%10"BLIN_U".%09lu ", tp.tv_sec, tp.tv_nsec);
            need += t;
            if  (0 > (avail -= t)) avail = 0; else ptr += t;
        }
        t = vsnprintf(ptr, avail + 1, format, ap);
        need += t;
        if  (0 >= (avail -= t)) avail = 0; else ptr += t;
        if  (ctl & 8) {
#           define UPREFIX         "Unknown error"
#           define EBUFSIZE        11

            if  (avail-- > 0) *(ptr++) = ':';
            if  (avail-- > 0) *(ptr++) = ' ';
            need += 2;
            if  ((er <= sys_nerr) && (er >= 0)) {
                for (t = 0; sys_errlist[er][t]; ++t) {
                    if  (avail > 0) {
                        *(ptr++) = sys_errlist[er][t];
                        --avail;
                    }
                    ++need;
                }
            } else if (avail < (EBUFSIZE + strlen(UPREFIX))) {
                need += EBUFSIZE + strlen(UPREFIX);
                avail = 0;
                if  (er < 0) {
                    ++need;
                    --avail;
                }
            } else {
                char *e;
                int s;

                for (t = 0; UPREFIX[t]; ++t) {
                    if  (avail > 0) {
                        *(ptr++) = UPREFIX[t];
                        --avail;
                    }
                    ++need;
                }
                e = ptr + EBUFSIZE;
                need += EBUFSIZE;
                avail -= EBUFSIZE;
                if  (er < 0) {
                    s = -er;
                    ++e;
                    ++need;
                    --avail;
                } else {
                    s = er;
                }
                do *--e = "0123456789"[s % 10]; while (s /= 10);
                if (er < 0) *--e = '-';
                while (e > ptr) *--e = ' ';
                ptr += EBUFSIZE;
                if  (er < 0) ++ptr;
        }   }
        if  (control & BLIN_MODEOAU) {
            if  (ptr[-1] != '\n') control |= BLIN_MODEOLF;
        }
        if  (!(~control & (BLIN_MODEOCR | BLIN_MODPRNT))) {
            *(ptr++) = '\r';
            --avail;
            ++need;
        }
        if  (!(~control & (BLIN_MODEOLF | BLIN_MODPRNT))) {
            *(ptr++) = '\n';
            --avail;
            ++need;
        }
        va_end(ap);
    } while (avail <= 0);
    *(ptr++) = '\0';
    doout(ctl & 7, buf);
out:
    if  (buf != qq) free(buf);
    if  (blin_pamparam_internal.flags & BLIN_TRAC) {
        fprintf(stderr, "BLIN: -blin_cry1 %d %"BLIN_D"\n", ex, need);
    }
    blin_pamparam_internal.err = errno;
    errno = er;
    return(ex);
}

int
/*****************************************************************************************************
 *****************************************************************************************************/
blin_stateheader(blin_statectl staflag, const char actnames[][3]) {                              /****
 *****************************************************************************************************
 *****************************************************************************************************/
    int i, ex = EX_OK;
    u_int32_t ctl;
    ssize_t l, t;
    char *q, *p;
#ifdef __amd64__
    const char *header = "nst rs cl st --rg32-- ----offs----";
#else
    const char *header = "nst rs cl st --rg32-- --offs--";
#endif

    blin_init();
    if  (blin_pamparam_internal.flags & (BLIN_BIT0 << staflag.level)) {
        if  (blin_pamparam_internal.flags & BLIN_TRAC) {
            fprintf(stderr, "BLIN: +blin_stateheader %08X\n", staflag.statectl);
        }
        dotime(staflag.level);
        ctl = blin_pamparam_internal.w[staflag.level].what;
        l = strlen("control-");
        if  (!staflag.allign) {
            for (i = 0; i < 24; ++i) if (actnames[i][0] != '-') l += 3;
        } else {
            l += 3 * staflag.allign;
        }
        l += strlen(header);
        l += 3; /* for \r\n\0 */
            /*     "---"   */
        if  (!(q = calloc(1, l + 3))) {
            ifBLIN_QX0("malloc");
            ERROUT(EX_OSERR, ENOMEM);
        }
        t = l;
        strncpy(q, "control-", t);
            /*  q  '\0'  strncpy()   calloc() */
        p = index(q, 0);
        if  (0 >= (t = q + l - p)) {
            ifBLIN_QX0("strcat #0");
            ERROUT(EX_SOFTWARE, EDOOFUS);
        }
        if  (!staflag.allign) {
            for (i = 0; i < 24; ++i) {
                if  (actnames[i][0] != '-') {
                    *(p++) = ' ';
                    *(p++) = actnames[i][0];
                    *(p++) = actnames[i][1];
                    if  (0 >= (t -= 3)) {
                        ifBLIN_QX0("strcat #1 %d", i);
                        ERROUT(EX_SOFTWARE, EDOOFUS);
            }   }   }
        } else  {
            for (i = 0; i < staflag.allign; ++i) {
                *(p++) = ' ';
                *(p++) = '-';
                *(p++) = '-';
                if  (0 >= (t -= 3)) {
                    ifBLIN_QX0("strcat #2 %d", i);
                    ERROUT(EX_SOFTWARE, EDOOFUS);
        }   }   }
            /*    (++t)      */
        strncat(q, header, ++t);
        p = index(q, 0);
        if  (0 > (t = q + l - p)) {
            ifBLIN_QX0("strcat #3 real %"BLIN_D" > %"BLIN_D" alloc", BLIN_I(p - q), BLIN_I(l));
            ERROUT(EX_SOFTWARE, EDOOFUS);
        }
        if  (ctl & BLIN_MODEOAU) {
            if  (p[-1] != '\n') ctl |= BLIN_MODEOLF;
        }
        if  (!(~ctl & (BLIN_MODEOCR | BLIN_MODPRNT))) {
            *(p++) = '\r';
            if  (0 >= (--t)) {
                ifBLIN_QX0("strcat #4");
                ERROUT(EX_SOFTWARE, EDOOFUS);
        }   }
        if  (!(~ctl & (BLIN_MODEOLF | BLIN_MODPRNT))) {
            *(p++) = '\n';
            if  (0 >= (--t)) {
                ifBLIN_QX0("strcat #5");
                ERROUT(EX_SOFTWARE, EDOOFUS);
        }   }
        *(p++) = '\0';
        doout(staflag.level, q);
out:
        if  (q) free(q);
        if  (blin_pamparam_internal.flags & BLIN_TRAC) {
            fprintf(stderr, "BLIN: -blin_stateheader %d\n", ex);
    }   }
    return(ex);
}

int
/*****************************************************************************************************
 *****************************************************************************************************/
blin_statebody( blin_statectl  staflag         /* level gtequ staten actinn bordsy allign revert     */
              , const char     actnames[][3]   /*                               */
              , const char     statenames[][3] /*                              */
              , const char     clasnname[3]    /*                            */
              , const char    *in              /*                        */
              , size_t         offset          /* 32/48 , ,    */
              , u_int64_t      control         /*                            */
              , u_int32_t      state           /*                                   */
              , u_int32_t      reg             /*   32                  */
              , u_char         byte            /*   8                   */
              ) {                                                                                   /*
 *****************************************************************************************************
 *****************************************************************************************************/
    u_int64_t C_state = BLIN_MIMASK(staflag.staten);
    u_int64_t flymask, contrana;
    int i, k, ex = EX_OK;
    u_int32_t ctl;
    ssize_t l, t;
    char *q, *p;

    blin_init();
    if  (blin_pamparam_internal.flags & (BLIN_BIT0 << staflag.level)) {
        if  (blin_pamparam_internal.flags & BLIN_TRAC) {
            fprintf(stderr, "BLIN: +blin_statebody %08X\n", staflag.statectl);
        }
        if  (staflag.revert) {
            flymask = (u_int64_t)1 << (staflag.staten + staflag.actinn - 1);
        } else {
            flymask = (u_int64_t)1 << (staflag.staten);
        }
        ctl = blin_pamparam_internal.w[staflag.level].what;
        l = 9;
        if  (!staflag.allign) {
            for (i = 0; i < 24; ++i) if (actnames[i][0] != '-') l += 3;
        } else {
            l += 3 * staflag.allign;
        }
        l += 3 + 3 + 3 + 3 + 9;
#ifdef __amd64__
        l += 13;
#else
        l += 9;
#endif
        l += BLIN_MAXTXTSTA + 3;
        l += 3; /* for \r\n\0 */
        if  (!(q = calloc(1, l + 3))) {
            ifBLIN_QX0("malloc");
            ERROUT(EX_OSERR, ENOMEM);
        }
        p = q + snprintf(q, l, "%08X ", (u_int32_t)control);
        if  (0 >= (t = q + l - p)) {
            ifBLIN_QX0("strcat #0");
            ERROUT(EX_SOFTWARE, EDOOFUS);
        }
        if  (!staflag.allign) {
            for (i = 0, contrana = control, k = 0; i < staflag.actinn; ++i) {
                if  (actnames[i][0] == '-') {
                    if  (contrana & flymask) {
                        ++k;
                    }
                } else {
                    if  (contrana & flymask) {
                        *(p++) = actnames[i][0];
                        *(p++) = actnames[i][1];
                    } else {
                        *(p++) = ' ';
                        *(p++) = ' ';
                    }
                    if  (k && (i == (staflag.actinn - 1))) {
                        *(p++) = '-';
                    } else {
                        *(p++) = ' ';
                    }
                    if  (0 >= (t -= 3)) {
                        ifBLIN_QX0("strcat #1 %d", i);
                        ERROUT(EX_SOFTWARE, EDOOFUS);
                }   }
                if  (staflag.revert) contrana <<= 1; else contrana >>= 1;
            }
        } else {
            for (i = 0, contrana = control, k = 0; i < staflag.actinn; ++i) {
                if  (contrana & flymask) {
                    *(p++) = actnames[i][0];
                    *(p++) = actnames[i][1];
                    *(p++) = ' ';
                    if  (0 >= (t -= 3)) {
                        ifBLIN_QX0("strcat #1 %d", i);
                        ERROUT(EX_SOFTWARE, EDOOFUS);
                    }
                    k++;
                }
                if  (staflag.revert) contrana <<= 1; else contrana >>= 1;
            }
            for (; k < staflag.allign; ++k) {
                *(p++) = ' ';
                *(p++) = ' ';
                *(p++) = ' ';
                if  (0 >= (t -= 3)) {
                    ifBLIN_QX0("strcat #2 %d", i);
                    ERROUT(EX_SOFTWARE, EDOOFUS);
        }   }   }
        p += snprintf( p, t, "%s %02X %s %s %08X %"BLIN_X" ~"
                     , statenames[control & C_state]
                     , byte & 0x00FF
                     , clasnname
                     , statenames[state]
                     , reg
                     , BLIN_I(offset)
                     );
        if  (0 >= (t = q + l - p)) {
            ifBLIN_QX0("strcat #3");
            ERROUT(EX_SOFTWARE, EDOOFUS);
        }
        for (i = 0; (i <= BLIN_MAXTXTSTA) && in[i]; ++i) {
            if  ((staflag.gtequ) ? (in[i] < staflag.bordsy) : (in[i] == staflag.bordsy)) break;
            *(p++) = in[i];
            if  (0 > (--t)) {
                ifBLIN_QX0("strcat #4");
                ERROUT(EX_SOFTWARE, EDOOFUS);
        }   }
        if  (i <= BLIN_MAXTXTSTA) {
            *(p++) = '~';
            --t;
        } else if (0 > t) {
            ifBLIN_QX0("strcat #5");
            ERROUT(EX_SOFTWARE, EDOOFUS);
        }
        if  (ctl & BLIN_MODEOAU) {
            if  (p[-1] != '\n') ctl |= BLIN_MODEOLF;
        }
        if  (!(~ctl & (BLIN_MODEOCR | BLIN_MODPRNT))) {
            *(p++) = '\r';
            if  (0 >= (--t)) {
                ifBLIN_QX0("strcat #6");
                ERROUT(EX_SOFTWARE, EDOOFUS);
        }   }
        if  (!(~ctl & (BLIN_MODEOLF | BLIN_MODPRNT))) {
            *(p++) = '\n';
            if  (0 >= (--t)) {
                ifBLIN_QX0("strcat #7");
                ERROUT(EX_SOFTWARE, EDOOFUS);
        }   }
        *(p++) = '\0';
        doout(staflag.level, q);
out:
        if  (q) free(q);
        if  (blin_pamparam_internal.flags & BLIN_TRAC) {
            fprintf(stderr, "BLIN: -blin_statebody %d\n", ex);
    }   }
    return(ex);
}

static char
z(u_char u) {
    char s;

    if  (u < ' ') s = '.';
    else if ((u < 0x7F) || (u >= (u_char)'') || (u == (u_char)'') || (u == (u_char)'')) s = u;
    else s = '.';
    return(s);
}

int
/*****************************************************************************************************
 *****************************************************************************************************/
blin_dumb(int mode, const void *buf, ssize_t len, ...) {                                         /****
 *****************************************************************************************************
 *****************************************************************************************************/
    char q[BLIN_MAXTXTDUM + BLIN_MAXSTRDUM + 3], *p, *s;
    int i, ex = EX_OK;
    u_int32_t control;
    const u_char *c;
    va_list ap;
    ssize_t t;
    size_t l;

    if  (blin_pamparam_internal.flags & BLIN_TRAC) {
        fprintf(stderr, "BLIN: +blin_dumb %08X %"BLIN_D"\n", mode, len);
    }
    c = buf;
    dotime(mode & 7);
    control = blin_pamparam_internal.w[mode & 7].what;
    while (len > 0) {
        p = q;
        if  (control & BLIN_MODDSTR) {
            va_start(ap, len);
            s = va_arg(ap, char *);
            if  (BLIN_MAXSTRDUM < (l = strnlen(s, BLIN_MAXSTRDUM + 3))) {
                ifBLIN_QX0("Long label %u > %u", l, BLIN_MAXSTRDUM);
                ERROUT(EX_USAGE, EMSGSIZE);
            }
            bcopy(s, p, l);
            p += l;
            va_end(ap);
        }
        if  (0 >= (t = q + BLIN_MAXTXTDUM - p)) {
            ifBLIN_QX0("strcat label");
            ERROUT(EX_SOFTWARE, EDOOFUS);
        }
        if  (control & BLIN_MODDNON) {
        } else if (control & BLIN_MODDREL) {
            p += snprintf(p, BLIN_MAXTXTDUM, "%04X", (int)(c - (u_char*)buf));
        } else {
            p += snprintf(p, BLIN_MAXTXTDUM, "%"BLIN_X, BLIN_I(c));
        }
        if  (0 >= (t = q + BLIN_MAXTXTDUM - p)) {
            ifBLIN_QX0("strcat #0");
            ERROUT(EX_SOFTWARE, EDOOFUS);
        }
        for (i = 0; i < 16; ++i) {
            if  (i && !(i & 3)) {
                *(p++) = ' ';
                t--;
                if  (0 > t) {
                    ifBLIN_QX0("strcat #1 %d", i);
                    ERROUT(EX_SOFTWARE, EDOOFUS);
            }   }
            if  (i < len) {
                p += snprintf(p, t, " %02X", c[i]);
            } else if ((mode & BLIN_DU_TEXT) || (control & BLIN_MODDTXT)) {
                *(p++) = ' ';
                *(p++) = ' ';
                *(p++) = ' ';
            }
            if  (0 >= (t = q + BLIN_MAXTXTDUM - p)) {
                ifBLIN_QX0("strcat #2 %d", i);
                ERROUT(EX_SOFTWARE, EDOOFUS);
        }   }
        if  ((mode & BLIN_DU_TEXT) || (control & BLIN_MODDTXT)) {
            *(p++) = ' ';
            *(p++) = '|';
            if  (0 >= (t -= 2)) {
                ifBLIN_QX0("strcat #3");
                ERROUT(EX_SOFTWARE, EDOOFUS);
            }
            for (i = 0; i < 16; ++i) {
                if  (i < len) {
                    *(p++) = z(c[i]);
                    if  (0 >= (--t)) {
                        ifBLIN_QX0("strcat #4");
                        ERROUT(EX_SOFTWARE, EDOOFUS);
            }   }   }
            *(p++) = '|';
            if  (0 >= (--t)) {
                ifBLIN_QX0("strcat #5");
                ERROUT(EX_SOFTWARE, EDOOFUS);
        }   }
        if  (control & BLIN_MODEOAU) control |= BLIN_MODEOLF;
        if  (!(~control & (BLIN_MODEOCR | BLIN_MODPRNT))) {
            *(p++) = '\r';
            if  (0 >= (--t)) {
                ifBLIN_QX0("strcat #6");
                ERROUT(EX_SOFTWARE, EDOOFUS);
        }   }
        if  (!(~control & (BLIN_MODEOLF | BLIN_MODPRNT))) {
            *(p++) = '\n';
            if  (0 >= (--t)) {
                ifBLIN_QX0("strcat #7");
                ERROUT(EX_SOFTWARE, EDOOFUS);
        }   }
        *(p++) = '\0';
        --t;
        doout(mode & 7, q);
        if  (len >= 16) {
            c += 16;
            len -= 16;
        } else {
            c += len;
            len = 0;
    }   }
out:
    if  (blin_pamparam_internal.flags & BLIN_TRAC) {
        fprintf(stderr, "BLIN: -blin_dumb %d\n", ex);
    }
    return(ex);
}
